from euclidean.util import normalize_coefficients
from euclidean.exceptions import unexpected_type_error
from euclidean.approx import approx

from .cartesian import P2, V2


class Line:
    @staticmethod
    def ByPoints(p1, p2):
        dx = p2.x - p1.x
        dy = p2.y - p1.y
        c = dx * p1.y - dy * p1.x
        return Line(-dy, dx, c)

    _cx = property(lambda self: self._coeffs[0])
    _cy = property(lambda self: self._coeffs[1])
    _c = property(lambda self: self._coeffs[2])

    def __init__(self, cx, cy, c):
        """
        cx * x + cy * y = c
        Args:
            cx:
            cy:
            c:
        """
        self._coeffs = normalize_coefficients(cx, cy, c)

    def _equation(self, point):
        return self._cx * point.x + self._cy * point.y

    def normal(self):
        return V2(self._cx, self._cy).unit()

    def translate(self, vector):
        cx = self._cx * (self._c + self._cy * vector.y)
        cy = self._cy * (self._c + self._cx * vector.x)
        c = self._c ** 2
        c += self._cx * self._c * vector.x
        c += self._cy * self._c * vector.y
        c += self._cx * self._cy * vector.x * vector.y
        return Line(cx, cy, c)

    def x(self, y):
        """Find the corresponding x coordinate on the line for coordinate y

        Args:
            y:

        Returns:

        """
        if self._cx == 0:
            return None
        return (self._c - self._cy * y) / self._cx

    def y(self, x):
        """Find the corresponding y coordinate on the line for coordinate x

        Args:
            x:

        Returns:

        """
        if self._cy == 0:
            return None
        return (self._c - self._cx * x) / self._cy

    def perpendicular(self, test_point):
        """Find the line perpendicular to this line, passing through test_point.

        If test_point is on this line, the case is degenerate and returns None

        Args:
            test_point (P2):

        Returns:
            (Line):
        """
        if self.contains(test_point):
            return None
        return Line(
            -self._cy, self._cx, self._cx * test_point.y - self._cy * test_point.x
        )

    def parallel(self, test_point):
        """Find the line parallel to this line, passing through test_point.

        Args:
            test_point (P2):

        Returns:
            (Line):
        """
        return Line(
            self._cx, self._cy, self._cx * test_point.x + self._cy * test_point.y
        )

    def closest(self, test_point):
        """Find the point on the line closest to the test_point.

        This should be a point that forms a line segment with test point perpendicular to this line.

        Args:
            point (P2):

        Returns:
            (P2): a point on the line
        """
        perpendicular = self.perpendicular(test_point)
        return self.intersection(perpendicular) if perpendicular else test_point

    def contains(self, point, atol=1e-6):
        if isinstance(point, P2):
            return abs(self._equation(point) - self._c) <= atol
        return False

    def __eq__(self, other):
        if isinstance(other, Line):
            return self._coeffs == other._coeffs
        return NotImplemented

    def __ne__(self, other):
        return not self == other

    def is_parallel(self, line, atol=1e-6):
        if not isinstance(line, Line):
            raise unexpected_type_error("line", Line, line)
        my_normal = self.normal()
        mirror = -my_normal
        other_normal = line.normal()
        return my_normal.approx(other_normal, atol) or mirror.approx(other_normal, atol)

    def does_intersect(self, line, atol=1e-6):
        if self == line:
            raise ValueError("Test Lines are identical.")
        return not self.is_parallel(line, atol)

    def intersection(self, line):
        if not isinstance(line, Line):
            raise unexpected_type_error("line", Line, line)

        det = self._cx * line._cy - self._cy * line._cx
        if det == 0:
            return None
        return P2(
            (line._cy * self._c - self._cy * line._c) / det,
            (self._cx * line._c - line._cx * self._c) / det,
        )

    def on_side(self, point):
        if self._cy == 0:
            if point.x > self.x(0):
                return 1
            if point.x == self.x(0):
                return 0
            return -1

        if point.y > self.y(point.x):
            return 1
        if point.y < self.y(point.x):
            return -1
        return 0


class LineSegment:
    def __init__(self, p1, p2):
        if p1 == p2:
            raise ValueError("Points must be independent to define a line segment.")
        self._p1 = p1
        self._p2 = p2

    def vector(self):
        return self._p2 - self._p1

    def length(self):
        return self.vector().magnitude()

    def ordered(self):
        return (
            (self._p1, self._p2)
            if self._p1._coords < self._p2._coords
            else (self._p2, self._p1)
        )

    def translate(self, vector):
        return LineSegment(self._p1 + vector, self._p2 + vector)

    def center(self):
        return self._p1 + self.vector() / 2

    def rotate(self, radians, around_point=None):
        around_point = around_point if around_point else self.center()
        return LineSegment(
            self._p1.rotate(radians, around_point),
            self._p2.rotate(radians, around_point),
        )

    def contains(self, point, atol=1e-6):
        if not isinstance(point, P2):
            raise unexpected_type_error("point", P2, point)

        if approx(self._p1, point, atol):
            return True
        if approx(self._p2, point, atol):
            return True

        line_vector = self.vector()
        test_vector = point - self._p1
        if not line_vector.is_parallel(test_vector, atol):
            return False
        return 0 <= line_vector.dot(test_vector) <= line_vector.dot(line_vector)

    def __eq__(self, other):
        if isinstance(other, LineSegment):
            return self.ordered() == other.ordered()
        return NotImplemented

    def __ne__(self, other):
        return not self == other

    def line(self):
        return Line.ByPoints(self._p1, self._p2)

    def intersection(self, other):
        if not self.does_intersect(other):
            return None
        return self.line().intersection(other.line())

    def does_intersect(self, line_segment):
        if not isinstance(line_segment, LineSegment):
            raise unexpected_type_error("line_segment", LineSegment, line_segment)
        ccw1 = P2.CCW(self._p1, self._p2, line_segment._p1)
        ccw2 = P2.CCW(self._p1, self._p2, line_segment._p2)
        if ccw1 * ccw2 > 0:
            return False
        ccw1 = P2.CCW(line_segment._p1, line_segment._p2, self._p1)
        ccw2 = P2.CCW(line_segment._p1, line_segment._p2, self._p2)
        if ccw1 * ccw2 > 0:
            return False
        return True

    def __hash__(self):
        return hash(self.ordered())

    def __str__(self):
        return "(%s, %s)" % (self._p1, self._p2)

    def __repr__(self):
        return "LineSegment(%r, %r)" % (self._p1, self._p2)



import operator

from euclidean.R2.cartesian import P2
from euclidean.R2.line import Line


def convex_hull(point_cloud):
    if len(point_cloud) < 3:
        return None

    return _divide_and_conquer_convex_hull(sorted(point_cloud)).points()


def _jarvis_convex_hull(points):
    """Find the convex hull of a point cloud using the jarvis march algorithm.

    Notes:

        O(n * h) for n points in point cloud, h points in hull

    Args:
        points (List[P2]):

    Returns:
        (List[P2]):
    """
    count = len(points)
    hull = []
    last_idx = 0
    while True:
        hull.append(points[last_idx])

        next_idx = (last_idx + 1) % count
        for test_idx, test_point in enumerate(points):
            if P2.CCW(points[last_idx], test_point, points[next_idx]) > 0:
                next_idx = test_idx

        last_idx = next_idx

        if last_idx == 0:
            return hull


def _divide_and_conquer_convex_hull(points):
    """

    Notes:

        O(n * log(n))

    Args:
        points:

    Returns:

    """
    count = len(points)
    if count < 6:
        return Hull(_jarvis_convex_hull(points))

    midpoint = count // 2
    min_cloud, max_cloud = points[:midpoint], points[midpoint:]

    min_hull = _divide_and_conquer_convex_hull(min_cloud)
    max_hull = _divide_and_conquer_convex_hull(max_cloud)

    return __merge_convex_hulls(min_hull, max_hull)


class Hull:
    class Vertex:
        __slots__ = ("cw", "point", "ccw")

        def __init__(self, point):
            self.point = point

    def __init__(self, points):
        p_iter = iter(points)
        first = Hull.Vertex(next(p_iter))
        last = Hull.Vertex(next(p_iter))
        first.cw = first.ccw = last
        last.cw = last.ccw = first
        self.__min = first

        for point in p_iter:
            self.__append(point)

    def __append(self, point):
        first = self.__min
        last = self.__min.cw
        new_vertex = Hull.Vertex(point)
        first.cw = new_vertex
        new_vertex.ccw = first
        last.ccw = new_vertex
        new_vertex.cw = last

    def points(self):
        current = self.__min
        while True:
            yield current.point
            current = current.ccw
            if current == self.__min:
                break

    def min(self):
        return self.__min

    def max(self):
        current = self.__min
        while current.point < current.ccw.point:
            current = current.ccw
        return current


def __merge_convex_hulls(min_hull, max_hull):

    (min_h_max_vtx, max_h_max_vtx) = __tangent(min_hull, max_hull, Tangent.MAX)
    (min_h_min_vtx, max_h_min_vtx) = __tangent(min_hull, max_hull, Tangent.MIN)

    min_h_max_vtx.cw = max_h_max_vtx
    max_h_max_vtx.ccw = min_h_max_vtx

    min_h_min_vtx.ccw = max_h_min_vtx
    max_h_min_vtx.cw = min_h_min_vtx

    return min_hull


import enum


class Tangent(enum.Enum):
    MAX = 1
    MIN = -1


def __tangent(min_hull, max_hull, dir):
    min_hull_vtx = min_hull.max()
    max_hull_vtx = max_hull.min()

    op = operator.gt if dir == Tangent.MAX else operator.lt

    while True:
        test_line = Line.ByPoints(min_hull_vtx.point, max_hull_vtx.point)

        cmp_vtx = min_hull_vtx.ccw if dir == Tangent.MAX else min_hull_vtx.cw
        if op(test_line.on_side(cmp_vtx.point), 0):
            min_hull_vtx = cmp_vtx
            continue

        cmp_vtx = max_hull_vtx.cw if dir == Tangent.MAX else max_hull_vtx.ccw
        if op(test_line.on_side(cmp_vtx.point), 0):
            max_hull_vtx = cmp_vtx
            continue

        return min_hull_vtx, max_hull_vtx



from siquant import make, si


def rotate(self, angle):
    return make(self.quantity.rotate(angle.get_as(si.radians)), self.units)


def unit(self):
    return make(self.quantity.unit(), si.unity)


def dot(self, other):
    return make(self.quantity.dot(other.quantity), self.units * other.units)


def cross(self, other):
    return make(self.quantity.cross(other.quantity), self.units * other.units)


def magnitude(self):
    return make(self.quantity.magnitude(), self.units)


def angle(self, other):
    if not hasattr(other, "quantity"):
        return make(self.quantity.angle(other), si.radians)
    return make(self.quantity.angle(other.quantity), si.radians)


def approx(self, other):
    return self.quantity.approx(other.quantity)


def is_parallel(self, other):
    if not hasattr(other, "quantity"):
        return self.quantity.is_parallel(other)
    return self.quantity.is_parallel(other.quantity)


def is_orthogonal(self, other):
    if not hasattr(other, "quantity"):
        return self.quantity.is_orthogonal(other)
    return self.quantity.is_orthogonal(other.quantity)


def manhattan_distance(self):
    return make(self.quantity.manhattan_distance(), self.units)



from euclidean.util import normalize_coefficients
from euclidean.exceptions import unexpected_type_error

from .cartesian import P3, V3


class Plane:

    cx = property(lambda self: self._coeffs[0])
    cy = property(lambda self: self._coeffs[1])
    cz = property(lambda self: self._coeffs[2])
    c = property(lambda self: self._coeffs[3])

    @staticmethod
    def PointNormal(point, normal_vector):
        c = (
            normal_vector.x * point.x
            + normal_vector.y * point.y
            + normal_vector.z * point.z
        )
        return Plane(normal_vector.x, normal_vector.y, normal_vector.z, c)

    def __init__(self, cx, cy, cz, c):
        self._coeffs = normalize_coefficients(cx, cy, cz, c)

    def _equation(self, point):
        if not isinstance(point, P3):
            raise unexpected_type_error("point", P3, point)
        return self.cx * point.x + self.cy * point.y + self.cz * point.z

    def normal(self):
        return V3(self.cx, self.cy, self.cz).unit()

    def contains(self, point, atol=1e-6):
        return abs(self._equation(point) - self.c) <= atol

    def __eq__(self, other):
        if not isinstance(other, Plane):
            return NotImplemented
        return self._coeffs == other._coeffs

    def __ne__(self, other):
        return not self == other



from itertools import chain

from euclidean.R2.cartesian import P2, V2, cross2
from euclidean.R2.line import LineSegment

from .hull import convex_hull
from .line_sweep import shamos_hoey


class Polygon:
    """

    """

    @classmethod
    def ConvexHull(cls, points):
        return cls(convex_hull(points), is_convex=True)

    def __init__(self, points, is_convex=None):
        self._points = tuple(points)
        if len(self._points) < 3:
            raise ValueError("At least 3 points are required to define a polygon.")
        self._min_index = _min_idx(self._points)
        self.__is_convex = is_convex

    def __len__(self):
        return len(self._points)

    def standard_form(self):
        """Normalize point order to begin traversal from minimum point.

        #todo: also detect if CW -> iterate backwards, ie. CCW?
        #todo: make this the default __iter__ method?

        Returns:

        """
        return self._rolled(self._min_index)

    def _rolled(self, offset):
        return _rolled(self._points, offset)

    def _cross_products(self):
        return map(cross2, self._points, self._rolled(1))

    def area(self):
        """Find the area of this polygon.

        Notes:

            This will return an incorrect value if the polygon is complex.

        Returns:

        """
        return 0.5 * abs(sum(self._cross_products()))

    def centroid(self):
        """Find the centroid of this polygon.

        Notes:

            This will return an incorrect value if the polygon is complex.

        Returns:

        """
        cx, cy, a = 0, 0, 0
        for p1, p2 in zip(self._points, self._rolled(1)):
            cross = cross2(p1, p2)
            cx += (p1.x + p2.x) * cross
            cy += (p1.y + p2.y) * cross
            a += cross
        a *= 3
        return P2(cx / a, cy / a)

    def translate(self, vector):
        return Polygon(p + vector for p in self._points)

    def centered_at(self, new_center_point):
        """Copy this polygon centered at the provided point.

        Returns:
            (Polygon):
        """
        vector = new_center_point - self.centroid()
        return Polygon(p + vector for p in self._points)

    def rotate(self, radians, center_point=None):
        """Rotate the polygon by radians around a center point or the centroid if none is provided.

        Args:
            radians:
            center_point:

        Returns:
            (Polygon)
        """
        center_point = center_point if center_point else self.centroid()
        return Polygon(p.rotate(radians, center_point) for p in self._points)

    def points(self):
        return self._points

    def xs(self):
        return (p.x for p in self._points)

    def ys(self):
        return (p.y for p in self._points)

    def __ccws(self):
        return map(P2.CCW, self._rolled(0), self._rolled(1), self._rolled(2))

    def __maybe_convex(self):
        return all(c <= 0 for c in self.__ccws()) or all(c >= 0 for c in self.__ccws())

    def is_convex(self):
        if self.__is_convex is None:
            self.__is_convex = len(self._points) < 4 or (
                self.__maybe_convex() and self.is_simple()
            )
        return self.__is_convex

    def is_simple(self):
        return shamos_hoey(self.edges())

    def edges(self):
        return map(LineSegment, self._points, chain(self._points[1:], self._points[:1]))

    def contains(self, test_point, atol=1e-6, closed=True):
        if self.winding_number(test_point) > 0:
            return True
        if closed:
            return self.on_perimeter(test_point, atol)
        return False

    def perimeter(self):
        return sum(edge.length() for edge in self.edges())

    def on_perimeter(self, point, atol=1e-6):
        return any(edge.contains(point, atol) for edge in self.edges())

    def winding_number(self, test_point):
        order = sum(self._cross_products())
        wn = 0
        for edge in self.edges():
            if edge._p1.y <= test_point.y:
                if edge._p2.y > test_point.y:
                    if order * P2.CCW(edge._p1, edge._p2, test_point) > 0:
                        wn += 1
            else:
                if edge._p2.y <= test_point.y:
                    if order * P2.CCW(edge._p1, edge._p2, test_point) < 0:
                        wn -= 1
        return wn

    def __eq__(self, other):
        if not isinstance(other, Polygon):
            return NotImplemented
        if len(self._points) != len(other._points):
            return False
        for p1, p2 in zip(self.standard_form(), other.standard_form()):
            if p1 != p2:
                return False
        return True

    def __ne__(self, other):
        return not self == other

    def plot(self, **kwargs):
        xs = list(self.xs())
        xs.append(xs[0])
        ys = list(self.ys())
        ys.append(ys[0])
        return plt.plot(xs, ys, **kwargs)


def _rolled(points, offset):
    return chain(points[offset:], points[:offset])


def _standard_form(points):
    return tuple(_rolled(points, _min_idx(points)))


def _min_idx(points):
    min_idx = 0
    for idx, point in enumerate(points):
        if point._coords < points[min_idx]._coords:
            min_idx = idx
    return min_idx



import math

from multipledispatch import dispatch

from euclidean.constants import tau

from euclidean.R2.cartesian import V2
from euclidean.R2.line import LineSegment
from euclidean.R2.circle import Circle
from euclidean.R2.polygon import Polygon


@dispatch(LineSegment)
def rasterize(line_segment, n=100):
    step_vec = line_segment.vector() / (n - 1)
    return [line_segment._p1 + i * step_vec for i in range(n)]


@dispatch(Circle)
def rasterize(circle, n=360):
    rot_theta = tau / n
    radius_vector = circle.radius * V2(1, 0)
    return [circle.center + radius_vector.rotate(rot_theta * i) for i in range(n)]


@dispatch(Polygon)
def rasterize(polygon, n=1000):
    n = int(n)
    assert n >= len(polygon)
    poly_points = list(polygon.points())
    poly_points.append(poly_points[0])
    return __rasterize_points(poly_points, n)


def __rasterize_points(points, n):
    print(points)

    count = len(points)
    if count == 2:
        (a, b) = points
        return __rasterize_pair(a, b, n)

    midpoint = count // 2
    n1 = n // 2
    raster = __rasterize_points(points[: midpoint + 1], n1)
    raster.extend(__rasterize_points(points[midpoint:], n - n1))
    return raster


def __rasterize_pair(p1, p2, n):
    assert n >= 1
    step_v = (p2 - p1) / n
    return [p1 + i * step_v for i in range(n)]



from multipledispatch import dispatch

from euclidean.R2.line import Line, LineSegment
from euclidean.R2.circle import Circle
from euclidean.R2.polygon import Polygon


@dispatch(Circle, Circle)
def does_intersect(circle1, circle2):
    return circle1.does_intersect(circle2)


@dispatch(Line, Line)
def does_intersect(line1, line2):
    return line1.does_intersect(line2)


@dispatch(LineSegment, LineSegment)
def does_intersect(line_segment1, line_segment2):
    return line_segment1.does_intersect(line_segment2)


@dispatch(Polygon, Polygon)
def does_intersect(polygon1, polygon2):
    # grab max and min points of polygons and return False if disjoint?

    for e1 in polygon1.edges():
        for e2 in polygon2.edges():
            if e1.does_intersect(e2):
                return True
    return False


@dispatch(Line, LineSegment)
def does_intersect(line, line_segment):
    return __does_intersect_line_line_segment(line, line_segment)


@dispatch(LineSegment, Line)
def does_intersect(line_segment, line):
    return __does_intersect_line_line_segment(line, line_segment)


@dispatch(Circle, Line)
def does_intersect(circle, line):
    return __does_intersect_circle_line(circle, line)


@dispatch(Line, Circle)
def does_intersect(line, circle):
    return __does_intersect_circle_line(circle, line)


@dispatch(Circle, LineSegment)
def does_intersect(circle, line_segment):
    return __does_intersect_circle_line_segment(circle, line_segment)


@dispatch(LineSegment, Circle)
def does_intersect(line_segment, circle):
    return __does_intersect_circle_line_segment(circle, line_segment)


@dispatch(Polygon, Line)
def does_intersect(polygon, line):
    return __does_intersect_polygon_line_like(polygon, line)


@dispatch(Line, Polygon)
def does_intersect(line, polygon):
    return __does_intersect_polygon_line_like(polygon, line)


@dispatch(Polygon, LineSegment)
def does_intersect(polygon, line_segment):
    return __does_intersect_polygon_line_like(polygon, line_segment)


@dispatch(LineSegment, Polygon)
def does_intersect(line_segment, polygon):
    return __does_intersect_polygon_line_like(polygon, line_segment)


def __does_intersect_circle_line(circle, line):
    eq = abs(line._equation(circle.center) - line._c)
    dist = eq / (line._cx ** 2 + line._cy ** 2) ** 0.5
    return dist <= circle.radius


def __does_intersect_circle_line_segment(circle, line_segment):
    closest = line_segment.line().closest(circle.center)
    p_min, p_max = line_segment.ordered()
    if closest > p_max:
        closest = p_max
    elif closest < p_min:
        closest = p_min

    dist_vec = circle.center - closest
    if dist_vec.dot(dist_vec) > circle.radius ** 2:
        return False
    return not circle.contains(p_min) or not circle.contains(p_max)


def __does_intersect_polygon_line_like(polygon, line):
    return any(does_intersect(line, edge) for edge in polygon.edges())


def __does_intersect_line_line_segment(line, line_segment):
    s1 = line.on_side(line_segment._p1)
    s2 = line.on_side(line_segment._p2)
    if s1 == 0 and s2 == 0:
        raise ValueError("Line Segment is a segment of Line.")
    return s1 != s2



import numbers
import math

from functools import total_ordering

from euclidean.exceptions import unexpected_type_error


def cross2(v1, v2):
    return v1.x * v2.y - v1.y * v2.x


def dot2(v1, v2):
    return v1.x * v2.x + v1.y * v2.y


class Cartesian2:
    __slots__ = ("_coords",)

    coords = property(lambda self: self._coords)

    x = property(lambda self: self._coords[0])
    y = property(lambda self: self._coords[1])

    @classmethod
    def Polar(cls, r, theta):
        return cls(r * math.cos(theta), r * math.sin(theta))

    def __init__(self, x, y):
        self._coords = (x, y)

    def __iter__(self):
        return iter(self._coords)

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self._coords == other._coords
        return NotImplemented

    def __hash__(self):
        return hash((self.__class__.__name__, self._coords))

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.x, self.y)


class V2(Cartesian2):
    __slots__ = ()

    def unit(self):
        return self / self.magnitude()

    def dot(self, other):
        if not isinstance(other, V2):
            raise unexpected_type_error("other", V2, other)
        return dot2(self, other)

    def cross(self, other):
        if not isinstance(other, V2):
            raise unexpected_type_error("other", V2, other)
        return cross2(self, other)

    def is_parallel(self, other, atol=1e-6):
        return abs(self.cross(other)) <= atol

    def is_orthogonal(self, other, atol=1e-6):
        return abs(self.dot(other)) <= atol

    def magnitude(self):
        return self.dot(self) ** 0.5

    def manhattan_distance(self):
        return self.x + self.y

    __abs__ = magnitude

    def angle(self, other):
        return math.acos(self.dot(other) / self.magnitude() / other.magnitude())

    def rotate(self, radians):
        cos = math.cos(radians)
        sin = math.sin(radians)
        return V2(self.x * cos - self.y * sin, self.x * sin + self.y * cos)

    def approx(self, other, atol=1e-6):
        if not isinstance(other, V2):
            raise unexpected_type_error("other", V2, other)
        return abs(self - other) < atol

    def __add__(self, other):
        if isinstance(other, V2):
            return V2(self.x + other.x, self.y + other.y)
        return NotImplemented

    __iadd__ = __add__

    def __sub__(self, other):
        if isinstance(other, V2):
            return V2(self.x - other.x, self.y - other.y)
        return NotImplemented

    __isub__ = __sub__

    def __mul__(self, other):
        if not isinstance(other, numbers.Real):
            return NotImplemented
        return V2(self.x * other, self.y * other)

    __imul__ = __rmul__ = __mul__

    def __truediv__(self, other):
        if not isinstance(other, numbers.Real):
            return NotImplemented
        return V2(self.x / other, self.y / other)

    __itruediv__ = __truediv__

    def __floordiv__(self, other):
        if not isinstance(other, numbers.Real):
            return NotImplemented
        return V2(self.x // other, self.y // other)

    __ifloordiv__ = __floordiv__

    def __neg__(self):
        return V2(-self.x, -self.y)


@total_ordering
class P2(Cartesian2):
    __slots__ = ()

    @staticmethod
    def CCW(p1, p2, p3):
        return (p2.x - p1.x) * (p3.y - p1.y) - (p3.x - p1.x) * (p2.y - p1.y)

    @staticmethod
    def AcuteAngle(p1, p2, p3):
        v1 = p1 - p2
        v2 = p3 - p2
        return v1.angle(v2)

    def vector(self):
        return V2(self.x, self.y)

    def rotate(self, radians, center_point=None):
        center_point = center_point if center_point else P2(0, 0)
        vector = self - center_point
        return center_point + vector.rotate(radians)

    def quadrant(self):
        if self.x >= 0 and self.y >= 0:
            return 1
        if self.x <= 0 and self.y >= 0:
            return 2
        if self.x <= 0 and self.y <= 0:
            return 3
        return 4

    def approx(self, other, atol=1e-6):
        if not isinstance(other, P2):
            raise unexpected_type_error("other", P2, other)
        return (self - other).magnitude() < atol

    def __lt__(self, other):
        if not isinstance(other, P2):
            return NotImplemented
        return self._coords < other._coords

    def __add__(self, other):
        if isinstance(other, V2):
            return P2(self.x + other.x, self.y + other.y)
        return NotImplemented

    __radd__ = __iadd__ = __add__

    def __sub__(self, other):
        if isinstance(other, V2):
            return P2(self.x - other.x, self.y - other.y)
        if isinstance(other, P2):
            return V2(self.x - other.x, self.y - other.y)
        return NotImplemented

    __isub__ = __sub__



from collections import deque


def normalize_coefficients(*coefficients):
    """Normalize coefficients such that all equivalents can be directly compared.

    :raises ValueError: If all coefficients are 0.

    :param *coefficients: The unnormalized coefficients.
    :rtype: Tuple[float, ...]
    """
    divisor = 0
    normalized = deque()
    for c in reversed(coefficients):
        if divisor:
            normalized.appendleft(c / divisor)
        elif c != 0:
            divisor = c
            normalized.appendleft(1)
        else:
            normalized.appendleft(0)

    if divisor == 0:
        raise ValueError("All coefficients are 0.")

    return tuple(normalized)


def methods(*meths):
    """A class decorator for attaching generic methods to a class.

    .. note:

        The first argument of the function will be bound to self instance.

    :param *meths: Raw functions to be bound to a class.
    """

    def decorator(cls):
        for meth in meths:
            setattr(cls, meth.__name__, meth)
        return cls

    return decorator



from .cartesian import V2, P2
from .circle import Circle
from .line import Line, LineSegment
from .polygon.polygon import Polygon
from .generic.intersect import intersect
from .generic.rasterize import rasterize
from .generic.does_intersect import does_intersect

__all__ = (
    "intersect",
    "does_intersect",
    "rasterize",
    "P2",
    "V2",
    "Line",
    "LineSegment",
    "Circle",
    "Polygon",
)



def unexpected_type_error(arg_name, expected_type, actual_value):
    return TypeError(
        "'{name}' must be {type!r} (got {value!r} that is a "
        "{actual!r}).".format(
            name=arg_name,
            type=expected_type,
            actual=actual_value.__class__,
            value=actual_value,
        ),
        arg_name,
        expected_type,
        actual_value,
    )



from sortedcontainers import SortedListWithKey


def shamos_hoey(edges):
    """Determine if a polyline is self intersecting.

    Args:
        edges:

    Returns:

    """
    POINT, EDGE, IDX, SIDE = 0, 1, 2, 3
    LEFT, RIGHT = -1, 1
    event_queue = SortedListWithKey(key=lambda event: event[POINT]._coords)
    for (idx, edge) in enumerate(edges):
        p_left, p_right = edge.ordered()
        e_left, e_right = (p_left, edge, idx, LEFT), (p_right, edge, idx, RIGHT)
        event_queue.add(e_left)
        event_queue.add(e_right)

    count = len(event_queue) / 2

    sweep_line = SortedListWithKey(
        key=lambda event: tuple(p._coords for p in event[EDGE].ordered())
    )

    def consecutive(e1, e2):
        distance = abs(e1[IDX] - e2[IDX])
        if 1 < distance < count - 1:
            return False
        return True

    while event_queue:
        event = event_queue.pop(0)
        if event[SIDE] == LEFT:
            sweep_line.add(event)
            idx = sweep_line.index(event)
            if idx + 1 < len(sweep_line):
                above = sweep_line[idx + 1]
                if not consecutive(event, above):
                    if event[EDGE].does_intersect(above[EDGE]):
                        return False
            if idx > 0:
                below = sweep_line[idx - 1]
                if not consecutive(event, below) and event[EDGE].does_intersect(
                    below[EDGE]
                ):
                    return False
        else:
            idx = sweep_line.bisect_left(event)
            if 0 < idx < len(sweep_line) - 1:
                above = sweep_line[idx]
                below = sweep_line[idx - 1]
                if not consecutive(above, below) and above[EDGE].does_intersect(
                    below[EDGE]
                ):
                    return False
            sweep_line.pop(idx)

    return True



from euclidean.constants import pi
from euclidean.exceptions import unexpected_type_error

from .cartesian import P2, V2


class Circle:

    center = property(lambda self: self._center)
    radius = property(lambda self: self._radius)
    diameter = property(lambda self: 2 * self._radius)

    def __init__(self, radius, center=P2(0, 0)):
        """

        (x - center.x) ** 2 + (y - center.y) ** 2 = radius ** 2

        Args:
            radius:
            center (P2):
        """

        if radius <= 0:
            raise ValueError("Radius must be positive.")
        self._center = center
        self._radius = radius

    def _equation(self, point):
        diff = point - self.center
        return diff.dot(diff)

    def on_circumference(self, point, atol=1e-6):
        return abs(self._equation(point) - self.radius ** 2) <= atol

    def circumference(self):
        return pi * self.diameter

    def area(self):
        return pi * self._radius ** 2

    def does_intersect(self, circle):
        if not isinstance(circle, Circle):
            raise unexpected_type_error("circle", Circle, circle)

        if self == circle:
            raise ValueError("Test circles are identical.")

        vector = circle.center - self.center
        magnitude = vector.magnitude()
        if magnitude > self.radius + circle.radius:
            return False
        if magnitude < abs(self.radius - circle.radius):
            return False

        return True

    def intersection(self, circle):
        if not self.does_intersect(circle):
            return set()

        my_c = self.center
        c_c = circle.center

        vector = c_c - my_c
        v_mag = vector.magnitude()

        apothem = (self.radius ** 2 - circle.radius ** 2 + v_mag ** 2) / (2 * v_mag)
        intersect_center = my_c + apothem * vector.unit()
        if apothem == self.radius:
            return {intersect_center}

        h = (self.radius ** 2 - apothem ** 2) ** 0.5
        offset = h / v_mag * V2(c_c.y - my_c.y, my_c.x - c_c.x)

        return {intersect_center + offset, intersect_center - offset}

    def contains(self, point):
        if isinstance(point, P2):
            return self._equation(point) <= self.radius ** 2
        return False



from multipledispatch import dispatch

from euclidean.R2.line import Line, LineSegment
from euclidean.R2.circle import Circle
from euclidean.R2.polygon import Polygon
from euclidean.R2.cartesian import P2, V2


@dispatch(Line, Line)
def intersect(line_1, line_2):
    point = line_1.intersection(line_2)
    return {point} if point else set()


@dispatch(LineSegment, LineSegment)
def intersect(line_segment_1, line_segment_2):
    point = line_segment_1.intersection(line_segment_2)
    return {point} if point else set()


@dispatch(Line, LineSegment)
def intersect(line, line_segment):
    return line_line_segment_intersection(line, line_segment)


@dispatch(LineSegment, Line)
def intersect(line_segment, line):
    return line_line_segment_intersection(line, line_segment)


@dispatch(Line, Circle)
def intersect(line, circle):
    return line_circle_intersection(line, circle)


@dispatch(Circle, Line)
def intersect(circle, line):
    return line_circle_intersection(line, circle)


@dispatch(LineSegment, Circle)
def intersect(line_segment, circle):
    return line_segment_circle_intersection(line_segment, circle)


@dispatch(Circle, LineSegment)
def intersect(circle, line_segment):
    return line_segment_circle_intersection(line_segment, circle)


@dispatch(Circle, Circle)
def intersect(circle_1, circle_2):
    return circle_circle_intersection(circle_1, circle_2)


@dispatch(Polygon, Line)
def intersect(polygon, line):
    return polygon_intersection(polygon, line)


@dispatch(Line, Polygon)
def intersect(line, polygon):
    return polygon_intersection(polygon, line)


@dispatch(Polygon, LineSegment)
def intersect(polygon, line_segment):
    return polygon_intersection(polygon, line_segment)


@dispatch(LineSegment, Polygon)
def intersect(line_segment, polygon):
    return polygon_intersection(polygon, line_segment)


@dispatch(Polygon, Circle)
def intersect(polygon, circle):
    raise NotImplementedError("todo")


@dispatch(Circle, Polygon)
def intersect(circle, polygon):
    raise NotImplementedError("todo")


# implementations


def line_line_segment_intersection(line, line_segment):
    point = line.intersection(line_segment.line())
    if line_segment.contains(point):
        return {point}
    return set()


def line_circle_intersection(line, circle):
    vector = circle.center.vector()
    line = line.translate(-vector)
    if line._cy == 0:
        p1 = P2(line.x(0), 0)
        p2 = P2(line.x(1), 1)
    else:
        p1 = P2(0, line.y(0))
        p2 = P2(1, line.y(1))

    dv = p2 - p1
    dr2 = dv.magnitude() ** 2
    determinant = p1.x * p2.y - p1.y * p2.x
    discriminant = circle.radius ** 2 * dr2 - determinant ** 2
    if discriminant < 0:
        return set()

    x0 = determinant * dv.y / dr2
    y0 = -determinant * dv.x / dr2

    central = P2(x0, y0) + vector

    if discriminant == 0:
        return {central}

    sqrt_descriminant = discriminant ** 0.5
    sign_y = -1 if dv.y < 0 else 1
    x_off = sign_y * dv.x * sqrt_descriminant / dr2
    y_off = abs(dv.y) * sqrt_descriminant / dr2
    offset = V2(x_off, y_off)

    return {central + offset, central - offset}


def line_segment_circle_intersection(line_segment, circle):
    points = intersect(line_segment.line(), circle)
    return set(p for p in points if line_segment.contains(p))


def circle_circle_intersection(circle_1, circle_2):
    return circle_1.intersection(circle_2)


def polygon_intersection(polygon, other):
    points = set()

    for edge in polygon.edges():
        points = points | intersect(edge, other)

    return points



from sortedcontainers import SortedList
from collections.abc import Iterable, Set


def approx(a, b, atol=1e-6):
    try:
        return abs(a - b) <= atol
    except TypeError:
        return False


class ApproxSet(Set):
    """A Set of approximable instances.

    .. note::

        In order for an item to be retained by this collection, it must support
        a **total ordering**. Rather than hashing values for membership, items
        are maintained in a sorted ordering to support log(n) membership testing.

    :param iterable:
    :param approx_fcn:

    """

    def __init__(self, iterable, approx_fcn=approx):
        self.__impl = SortedList()
        self.__approx = approx_fcn
        for item in iterable:
            if item not in self:
                self.__impl.add(item)

    def __len__(self):
        return len(self.__impl)

    def __iter__(self):
        return iter(self.__impl)

    def __eq__(self, other):
        if isinstance(other, ApproxSet):
            return self.__equals(other)
        if isinstance(other, Set):
            return self.__equals(other)
        if isinstance(other, Iterable):
            return self.__equals(ApproxSet(other))
        return NotImplemented

    def __equals(self, sequence):
        if len(self) != len(sequence):
            return False
        sequence = sequence if isinstance(sequence, ApproxSet) else ApproxSet(sequence)
        return all(self.__approx(a, b) for a, b in zip(self.__impl, sequence))

    def __contains__(self, item):
        """Does the Set contain an item approximately equal to ``item``.

        :param item: The item to test for membership.
        :rtype: ``bool``
        """
        idx = self.__impl.bisect_right(item)
        return self.__test_index(idx, item) or self.__test_index(idx - 1, item)

    def __test_index(self, index, item):
        try:
            return self.__approx(self.__impl[index], item)
        except IndexError:
            return False



import math
import numbers

from functools import total_ordering

from euclidean.R2.cartesian import V2
from euclidean.exceptions import unexpected_type_error


class Cartesian3:
    __slots__ = ("_coords",)

    def __init__(self, x, y, z):
        self._coords = (x, y, z)

    x = property(lambda self: self._coords[0])
    y = property(lambda self: self._coords[1])
    z = property(lambda self: self._coords[2])

    def __iter__(self):
        return iter(self._coords)

    def __repr__(self):
        return "%s(%r, %r, %r)" % (self.__class__.__name__, self.x, self.y, self.z)

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self._coords == other._coords
        return NotImplemented


class V3(Cartesian3):
    __slots__ = ()

    def dot(self, other):
        if not isinstance(other, V3):
            raise unexpected_type_error("other", V3, other)
        return self.x * other.x + self.y * other.y + self.z * other.z

    def cross(self, other):
        return V3(
            self.y * other.z - self.z * other.y,
            self.z * other.x - self.x * other.z,
            self.x * other.y - self.y * other.x,
        )

    def magnitude(self):
        return self.dot(self) ** 0.5

    __abs__ = magnitude

    def unit(self):
        return self / self.magnitude()

    def angle(self, other):
        return math.acos(self.dot(other) / self.magnitude() / other.magnitude())

    def xy(self):
        return V2(self.x, self.y)

    def xz(self):
        return V2(self.x, self.z)

    def yz(self):
        return V2(self.y, self.z)

    def manhattan_distance(self):
        return sum(self._coords)

    def approx(self, other, atol=1e-6):
        if not isinstance(other, V3):
            raise unexpected_type_error("other", V3, other)
        return abs(self - other) <= atol

    def __add__(self, other):
        if isinstance(other, V3):
            return V3(self.x + other.x, self.y + other.y, self.z + other.z)
        return NotImplemented

    __iadd__ = __add__

    def __sub__(self, other):
        if isinstance(other, V3):
            return V3(self.x - other.x, self.y - other.y, self.z - other.z)
        return NotImplemented

    __isub__ = __sub__

    def __mul__(self, other):
        if isinstance(other, numbers.Real):
            return V3(self.x * other, self.y * other, self.z * other)
        return NotImplemented

    __imul__ = __rmul__ = __mul__

    def __truediv__(self, other):
        if isinstance(other, numbers.Real):
            return V3(self.x / other, self.y / other, self.z / other)
        return NotImplemented

    __itruediv__ = __truediv__

    def __floordiv__(self, other):
        if isinstance(other, numbers.Real):
            return V3(self.x // other, self.y // other, self.z // other)
        return NotImplemented

    def __neg__(self):
        return V3(-self.x, -self.y, -self.z)


@total_ordering
class P3(Cartesian3):
    __slots__ = ()

    def __add__(self, other):
        if isinstance(other, V3):
            return P3(self.x + other.x, self.y + other.y, self.z + other.z)
        return NotImplemented

    __radd__ = __iadd__ = __add__

    def approx(self, other, atol=1e-6):
        if not isinstance(other, P3):
            raise unexpected_type_error("other", P3, other)
        return abs(self - other) < atol

    def __sub__(self, other):
        if isinstance(other, V3):
            return P3(self.x - other.x, self.y - other.y, self.z - other.z)
        if isinstance(other, P3):
            return V3(self.x - other.x, self.y - other.y, self.z - other.z)
        return NotImplemented

    def vector(self):
        return V3(self.x, self.y, self.z)

    def __lt__(self, other):
        if not isinstance(other, P3):
            return NotImplemented
        return self._coords < other._coords



from setuptools import setup, find_packages


def readme():
    with open("README.rst", "r") as f:
        return f.read()


CLASSIFIERS = []
KEYWORDS = ["euclidean", "geometry", "vector", "point", "polygon"]
PROJECT_URLS = {}


INSTALL_REQUIRES = ["sortedcontainers>=2", "multipledispatch>=0.6"]
TESTS_REQUIRE = ["pytest", "coverage"]
EXTRAS_REQUIRE = {
    "docs": [],
    "tests": TESTS_REQUIRE,
    "siquant": ["siquant==4.0.0b6"],
    "dev": TESTS_REQUIRE + ["pre-commit"],
}

setup(
    name="euclidean",
    version="1.0.0b3",
    description="Euclidean Geometry Library",
    long_description=readme(),
    url="https://github.com/keystonetowersystems/euclidean",
    project_urls=PROJECT_URLS,
    author="Greg Echelberger",
    author_email="greg@keystonetowersystems.com",
    packages=find_packages(exclude=("tests",)),
    setup_requires=[],
    extras_require=EXTRAS_REQUIRE,
    install_requires=INSTALL_REQUIRES,
    zip_safe=True,
)



from siquant import Quantity

import euclidean.siquant.vector as v

from euclidean.R2.cartesian import V2
from euclidean.R3.cartesian import V3
from euclidean.util import methods


def make_factory(*quantity_types):
    def factory(q, u):
        for q_type in quantity_types:
            if isinstance(q, q_type._TYPE):
                return q_type(q, u)
        return Quantity(q, u)

    return staticmethod(factory)


@methods(
    v.angle,
    v.unit,
    v.rotate,
    v.magnitude,
    v.approx,
    v.cross,
    v.dot,
    v.is_parallel,
    v.is_orthogonal,
    v.manhattan_distance,
)
class V2Quantity(Quantity):
    __slots__ = ()
    _TYPE = V2

    x = property(lambda self: self.quantity.x * self.units)
    y = property(lambda self: self.quantity.y * self.units)


@methods(
    v.angle,
    v.unit,
    v.magnitude,
    v.approx,
    v.cross,
    v.dot,
    v.is_parallel,
    v.is_orthogonal,
    v.manhattan_distance,
)
class V3Quantity(Quantity):
    __slots__ = ()

    _TYPE = V3

    x = property(lambda self: self.quantity.x * self.units)
    y = property(lambda self: self.quantity.y * self.units)
    z = property(lambda self: self.quantity.z * self.units)

    def xy(self):
        return self.quantity.xy() * self.units

    def xz(self):
        return self.quantity.xz() * self.units

    def yz(self):
        return self.quantity.yz() * self.units
